from typing import Optional

from naneos.partector.blueprints._data_structure import NaneosDeviceDataPoint
from naneos.partector_ble.decoder.partector_ble_decoder_blueprint import (
    PartectorBleDecoderBlueprint,
)


class PartectorBleDecoderStd(PartectorBleDecoderBlueprint):
    """
    Decode the std advertisement data from the Partector device.
    """

    OFFSET_SERIAL_NUMBER = slice(14, 16)

    OFFSET_DEVICE_STATE_1 = slice(10, 12)
    ELEMET_DEVICE_STATE_2 = 19

    OFFSET_LDSA = slice(0, 3)
    OFFSET_AVERAGE_PARTICLE_DIAMETER = slice(3, 5)
    OFFSET_PARTICLE_NUMBER = slice(5, 8)
    OFFSET_TEMPERATURE = slice(8, 9)
    OFFSET_RELATIVE_HUMIDITY = slice(9, 10)
    OFFSET_BATTERY_VOLTAGE = slice(12, 14)
    OFFSET_PARTICLE_MASS = slice(14, 18)

    FACTOR_LDSA = 0.01
    FACTOR_BATTERY_VOLTAGE = 0.01
    FACTOR_PARTICLE_MASS = 0.01

    # == External used methods =====================================================================
    @classmethod
    def get_serial_number(cls, data: bytes) -> int:
        """
        Get the serial number from the advertisement data.
        """
        return int.from_bytes(data[cls.OFFSET_SERIAL_NUMBER], byteorder="little")

    @classmethod
    def decode(
        cls, data: bytes, data_structure: Optional[NaneosDeviceDataPoint] = None
    ) -> NaneosDeviceDataPoint:
        """
        Decode the advertisement data from the Partector device.
        """
        decoded_data = NaneosDeviceDataPoint(
            serial_number=cls.get_serial_number(data),
            ldsa=cls._get_ldsa(data),
            average_particle_diameter=cls._get_diameter(data),
            particle_number_concentration=cls._get_particle_number(data),
            temperature=cls._get_temperature(data),
            relative_humidity=cls._get_relative_humidity(data),
            device_status=cls._get_device_state(data),
            battery_voltage=cls._get_battery_voltage(data),
            particle_mass=cls._get_particle_mass(data),
        )

        if not data_structure:
            return decoded_data

        # Fill the given data_structure with the decoded data
        for field in NaneosDeviceDataPoint.BLE_STD_FIELD_NAMES:
            setattr(data_structure, field, getattr(decoded_data, field))
        return data_structure

    # == Helpers ===================================================================================
    @classmethod
    def _get_ldsa(cls, data: bytes) -> float:
        """
        Get the LDSA from the advertisement data.
        """
        val = float(int.from_bytes(data[cls.OFFSET_LDSA], byteorder="little"))
        return val * cls.FACTOR_LDSA

    @classmethod
    def _get_diameter(cls, data: bytes) -> float:
        """
        Get the particle diameter from the advertisement data.
        """
        val = float(int.from_bytes(data[cls.OFFSET_AVERAGE_PARTICLE_DIAMETER], byteorder="little"))
        return val

    @classmethod
    def _get_particle_number(cls, data: bytes) -> float:
        """
        Get the particle number from the advertisement data.
        """
        val = float(int.from_bytes(data[cls.OFFSET_PARTICLE_NUMBER], byteorder="little"))
        return val

    @classmethod
    def _get_temperature(cls, data: bytes) -> float:
        """
        Get the temperature from the advertisement data.
        """
        val = float(int.from_bytes(data[cls.OFFSET_TEMPERATURE], byteorder="little"))
        return val

    @classmethod
    def _get_relative_humidity(cls, data: bytes) -> float:
        """
        Get the relative humidity from the advertisement data.
        """
        val = float(int.from_bytes(data[cls.OFFSET_RELATIVE_HUMIDITY], byteorder="little"))
        return val

    @classmethod
    def _get_device_state(cls, data: bytes) -> int:
        """
        Get the device state from the advertisement data.
        """
        val = int.from_bytes(data[cls.OFFSET_DEVICE_STATE_1], byteorder="little")
        val += ((int(data[cls.ELEMET_DEVICE_STATE_2]) >> 1) & 0b01111111) << 16
        return val

    @classmethod
    def _get_battery_voltage(cls, data: bytes) -> float:
        """
        Get the battery voltage from the advertisement data.
        """
        val = float(int.from_bytes(data[cls.OFFSET_BATTERY_VOLTAGE], byteorder="little"))
        return val * cls.FACTOR_BATTERY_VOLTAGE

    @classmethod
    def _get_particle_mass(cls, data: bytes) -> float:
        """
        Get the particle mass from the advertisement data.
        """
        val = float(int.from_bytes(data[cls.OFFSET_PARTICLE_MASS], byteorder="little"))
        return val * cls.FACTOR_PARTICLE_MASS



import datetime as dt

import pandas as pd
from influxdb_client.client.influxdb_client import InfluxDBClient

from naneos.logger import get_naneos_logger

URL_INFLUX = "https://influxdb.naneos.ch"
ORG_INFLUX = "naneos"

logger = get_naneos_logger(__name__)


def get_query(bucket: str, serial_number: str, start: str, stop: str) -> str:
    query = f"""
        from(bucket: "{bucket}")
        |> range(start: {start}, stop: {stop})
        |> filter(fn: (r) => r["_measurement"] == "v6_sensor")
        |> filter(fn: (r) => r["serial_number"] == "{serial_number}")
        |> keep(columns: ["_time", "_value", "_field"])
        |> pivot(rowKey:["_time"], columnKey: ["_field"], valueColumn: "_value")
    """

    return query


def create_start_stop_timestamp(start_dt: dt.datetime, stop_dt: dt.datetime) -> list:
    """Returns a list of lists with start stop times in 1 day chunks in unix timesamps"""
    delta_seconds = 3600 * 24 * 2

    start = int(start_dt.timestamp())
    stop = int(stop_dt.timestamp())

    start_stop_times = []

    while start < stop - delta_seconds:
        start_stop_times.append([start, start + delta_seconds])
        start += delta_seconds
    start_stop_times.append([start, stop])

    return start_stop_times


def download_from_iotweb(
    name: str, serial_number: str, start: dt.datetime, stop: dt.datetime, token: str
) -> pd.DataFrame:
    """Download your data from influxdb.naneos.ch.
    1 Month of data takes about 30 seconds to download and uses about 100 MB of data.

    You need to have a token to access the data.
    Ask mario.huegi@naneos.ch for your read token.
    We kindly ask you to not overuse our server.
    If you need to download the same data in a recuring pattern, contact us.

    Args:
        name (str): Name of the influx bucket.
        serial_number (str): Serial number of your device as string.
        start (dt.datetime): Start date of the data you want to download.
        stop (dt.datetime): End date of the data you want to download.
        token (str): Your read token. Do not push your token to public repositories.

    Returns:
        pd.DataFrame: Dataframe with your data.
    """
    timestamps = create_start_stop_timestamp(start, stop)

    dfs = []

    with InfluxDBClient(url=URL_INFLUX, org=ORG_INFLUX, token=token) as client:
        for t1, t2 in timestamps:
            query = get_query(name, serial_number, t1, t2)

            df = client.query_api().query_data_frame(query)

            if isinstance(df, list):
                dfs.extend(df)
            elif isinstance(df, pd.DataFrame):
                dfs.append(df)
            else:
                logger.warning(f"Unknown type: {type(df)}")

    df = pd.concat(dfs, axis=0)
    df.set_index("_time", inplace=True)
    df.drop(["result", "table"], axis=1, inplace=True)

    return df


if __name__ == "__main__":
    import os

    token = os.getenv("IOT_GUEST_TOKEN", None)
    if token is None:
        raise ValueError("No token found in your environment")

    start = dt.datetime(2023, 11, 25)
    stop = dt.datetime(2023, 11, 27)
    serial_number = "8134"
    name = "iot_guest"

    df = download_from_iotweb(name, serial_number, start, stop, token)
    print(df)



import asyncio
import threading
import time
from typing import Dict

import pandas as pd
from bleak.backends.device import BLEDevice

from naneos.logger import LEVEL_WARNING, get_naneos_logger
from naneos.partector.blueprints._data_structure import (
    NaneosDeviceDataPoint,
)
from naneos.partector_ble.partector_ble_connection import PartectorBleConnection
from naneos.partector_ble.partector_ble_scanner import PartectorBleScanner

pd.set_option("future.no_silent_downcasting", True)

logger = get_naneos_logger(__name__, LEVEL_WARNING)


class PartectorBleManager(threading.Thread):
    def __init__(self) -> None:
        super().__init__(daemon=True)
        self._stop_event = threading.Event()

        self._queue_scanner = PartectorBleScanner.create_scanner_queue()
        self._queue_connection = PartectorBleConnection.create_connection_queue()
        self._connections: Dict[int, asyncio.Task] = {}  # key: serial_number

        self._data: dict[int, pd.DataFrame] = {}

    def get_data(self) -> dict[int, pd.DataFrame]:
        """Returns the data dictionary and deletes it."""
        data = self._data
        self._data = {}
        return data

    def stop(self) -> None:
        self._stop_event.set()

    def run(self) -> None:
        try:
            asyncio.run(self._async_run())
        except RuntimeError as e:
            logger.exception(f"BLEManager loop exited with: {e}")

    def get_connected_device_strings(self) -> list[str]:
        """Returns a list of connected device strings."""
        return [f"SN{sn}" for sn in self._connections.keys()]

    def get_connected_serial_numbers(self) -> list[int | None]:
        """Returns a list of connected serial numbers."""
        return list(self._connections.keys())

    async def _async_run(self):
        self._loop = asyncio.get_event_loop()
        try:
            async with PartectorBleScanner(loop=self._loop, queue=self._queue_scanner):
                logger.info("Scanner started.")
                await self._manager_loop()
        except asyncio.CancelledError:
            logger.info("BLEManager cancelled.")
        finally:
            logger.info("BLEManager cleanup complete.")

    async def _manager_loop(self) -> None:
        while not self._stop_event.is_set():
            try:
                await asyncio.sleep(1.0)

                await self._scanner_queue_routine()
                await self._connection_queue_routine()
                await self._remove_done_tasks()

            except asyncio.TimeoutError:
                continue
            except Exception as e:
                logger.exception(f"Error in scanner loop: {e}")

        # wait for all connections to finish
        for serial in list(self._connections.keys()):
            if not self._connections[serial].done():
                logger.info(f"Waiting for connection task {serial} to finish.")
                await self._connections[serial]
            self._connections.pop(serial, None)
            logger.info(f"{serial}: Connection task finished and popped.")

    async def _task_connection(self, device: BLEDevice, serial: int) -> None:
        try:
            async with PartectorBleConnection(
                device=device, loop=self._loop, serial_number=serial, queue=self._queue_connection
            ):
                while not self._stop_event.is_set():
                    await asyncio.sleep(1)

        except asyncio.CancelledError:
            logger.info(f"{serial}: Connection task cancelled.")
        except Exception as e:
            logger.warning(f"{serial}: Connection task failed: {e}")
        finally:
            logger.info(f"{serial}: Connection task finished.")

    async def _scanner_queue_routine(self) -> None:
        to_check: dict[int, BLEDevice] = {}

        while not self._queue_scanner.empty():
            device, decoded = await self._queue_scanner.get()
            if not decoded.serial_number:
                continue

            self._data = NaneosDeviceDataPoint.add_data_point_to_dict(self._data, decoded)
            to_check[decoded.serial_number] = device

        # check for new devices
        for serial, device in to_check.items():
            if serial in self._connections:
                continue  # already connected

            logger.info(f"New device detected: serial={serial}, address={device.address}")
            task = self._loop.create_task(self._task_connection(device, serial))
            self._connections[serial] = task

    async def _connection_queue_routine(self) -> None:
        while not self._queue_connection.empty():
            data = await self._queue_connection.get()
            self._data = NaneosDeviceDataPoint.add_data_point_to_dict(self._data, data)

    async def _remove_done_tasks(self) -> None:
        """Remove completed tasks from the connections dictionary."""
        for serial in list(self._connections.keys()):
            if self._connections[serial].done():
                self._connections.pop(serial, None)
                logger.info(f"{serial}: Connection task finished and popped.")


if __name__ == "__main__":
    manager = PartectorBleManager()
    manager.start()

    for _ in range(2):
        time.sleep(10)  # Allow some time for the scanner to start
        data = manager.get_data()

        print(f"Connected serial numbers: {manager.get_connected_serial_numbers()}")
        print("Collected data:")
        print()

        for sn, df in data.items():
            print(f"SN: {sn}")
            print(df)
            print("-" * 40)
            print()

    manager.stop()
    manager.join()



from typing import Optional

import pandas as pd

import naneos.protobuf.protoV1_pb2 as pbScheme
from naneos.partector.blueprints._data_structure import NaneosDeviceDataPoint


def create_combined_entry(
    devices: list[pbScheme.Device],
    abs_timestamp: int,
    gateway_points: Optional[list[pbScheme.GatewayPointLegacy]] = None,
    position_points: Optional[list[pbScheme.PositionPoint]] = None,
    wind_points: Optional[list[pbScheme.WindPoint]] = None,
) -> pbScheme.CombinedData:
    combined = pbScheme.CombinedData()
    combined.abs_timestamp = abs_timestamp

    combined.devices.extend(devices)

    if gateway_points is not None:
        combined.gateway_points_legacy.extend(gateway_points)

    if position_points is not None:
        combined.position_points.extend(position_points)

    if wind_points is not None:
        combined.wind_points.extend(wind_points)

    return combined


def create_proto_device(sn: int, abs_time: int, df: pd.DataFrame) -> pbScheme.Device:
    device = pbScheme.Device()
    device.type = (
        int(df["device_type"].iloc[-1])
        if "device_type" in df
        else NaneosDeviceDataPoint.DEV_TYPE_P2
    )
    device.serial_number = sn

    device_points = df.apply(_create_device_point, axis=1, abs_time=abs_time).to_list()  # type: ignore
    device_points = [x for x in device_points if x is not None]

    device.device_points.extend(device_points)

    return device


def _create_device_point(ser: pd.Series, abs_time: int) -> Optional[pbScheme.DevicePoint]:
    try:
        device_point = pbScheme.DevicePoint()

        ser = ser.dropna()

        # mandatory fields
        if isinstance(ser.name, int):
            timestamp = ser.name
        else:
            raise ValueError("Timestamp is not an int!")
        device_point.timestamp = abs_time - timestamp
        if "device_status" in ser:
            device_point.device_status = int(ser["device_status"])
        if "firmware_version" in ser:
            device_point.firmware_version = int(ser["firmware_version"])

        if "ldsa" in ser:
            device_point.ldsa = int(ser["ldsa"] * 100.0)
        if "particle_number_concentration" in ser:
            device_point.particle_number_concentration = int(
                round(ser["particle_number_concentration"])
            )
        if "average_particle_diameter" in ser:
            device_point.average_particle_diameter = int(round(ser["average_particle_diameter"]))
        if "particle_mass" in ser:
            device_point.particle_mass = int(round(ser["particle_mass"] * 100.0))
        if "particle_surface" in ser:
            device_point.particle_surface = int(round(ser["particle_surface"] * 100.0))
        if "diffusion_current" in ser:
            idiff_tmp = ser["diffusion_current"] if ser["diffusion_current"] > 0 else 0
            device_point.diffusion_current = int(round(idiff_tmp * 100.0))
        if "diffusion_current_offset" in ser:
            device_point.diffusion_current_offset = int(
                round(ser["diffusion_current_offset"] * 100.0)
            )
        if "diffusion_current_stddev" in ser:
            device_point.diffusion_current_stddev = int(
                round(ser["diffusion_current_stddev"] * 100.0)
            )
        # TODO: implement in protobuf
        # if "diffusion_current_average" in ser:
        #     device_point.diffusion_current_average = int(
        #         round(ser["diffusion_current_average"] * 100.0)
        #     )
        # if "diffusion_current_max" in ser:
        #     device_point.diffusion_current_max = int(round(ser["diffusion_current_max"] * 100.0))
        if "diffusion_current_delay_on" in ser:
            delay_on_tmp = max(ser["diffusion_current_delay_on"], 0)
            device_point.diffusion_current_delay_on = int(round(delay_on_tmp))
        if "diffusion_current_delay_off" in ser:
            delay_off_tmp = max(ser["diffusion_current_delay_off"], 0)
            device_point.diffusion_current_delay_off = int(round(delay_off_tmp))
        if "corona_voltage" in ser:
            device_point.corona_voltage = int(round(ser["corona_voltage"]))
        # TODO: implement in protobuf
        # if "corona_voltage_onset" in ser:
        #     device_point.corona_voltage_onset = int(round(ser["corona_voltage_onset"]))
        if "electrometer_1_amplitude" in ser:
            device_point.electrometer_1_offset = int(round(ser["electrometer_1_amplitude"] * 10.0))
        if "electrometer_2_amplitude" in ser:
            device_point.electrometer_2_offset = int(round(ser["electrometer_2_amplitude"] * 10.0))
        if "electrometer_1_gain" in ser:
            device_point.electrometer_1_gain = int(round(ser["electrometer_1_gain"] * 100.0))
        if "electrometer_2_gain" in ser:
            device_point.electrometer_2_gain = int(round(ser["electrometer_2_gain"] * 100.0))
        if "temperature" in ser:
            device_point.temperature = int(round(ser["temperature"]))
        if "relative_humidity" in ser:
            device_point.relative_humidity = int(round(ser["relative_humidity"]))
        if "flow_from_dp" in ser:
            device_point.flow = int(round(ser["flow_from_dp"] * 1000.0))
        if "deposition_voltage" in ser:
            device_point.deposition_voltage = int(round(ser["deposition_voltage"]))
        if "battery_voltage" in ser:
            device_point.battery_voltage = int(round(ser["battery_voltage"] * 100.0))
        if "ambient_pressure" in ser:
            device_point.ambient_pressure = int(round(ser["ambient_pressure"] * 10.0))
        if "channel_pressure" in ser:
            device_point.channel_pressure = int(round(ser["channel_pressure"] * 10.0))
        if "differential_pressure" in ser:
            device_point.differential_pressure = int(round(ser["differential_pressure"] * 10.0))
        if "pump_voltage" in ser:
            device_point.pump_voltage = int(round(ser["pump_voltage"] * 100.0))
        if "pump_current" in ser:
            device_point.pump_current = int(round(ser["pump_current"] * 1000.0))
        if "pump_pwm" in ser:
            device_point.pump_pwm = int(round(ser["pump_pwm"]))
        if "particle_number_10nm" in ser:
            device_point.particle_number_10nm = int(round(ser["particle_number_10nm"]))
        if "particle_number_16nm" in ser:
            device_point.particle_number_16nm = int(round(ser["particle_number_16nm"]))
        if "particle_number_26nm" in ser:
            device_point.particle_number_26nm = int(round(ser["particle_number_26nm"]))
        if "particle_number_43nm" in ser:
            device_point.particle_number_43nm = int(round(ser["particle_number_43nm"]))
        if "particle_number_70nm" in ser:
            device_point.particle_number_70nm = int(round(ser["particle_number_70nm"]))
        if "particle_number_114nm" in ser:
            device_point.particle_number_114nm = int(round(ser["particle_number_114nm"]))
        if "particle_number_185nm" in ser:
            device_point.particle_number_185nm = int(round(ser["particle_number_185nm"]))
        if "particle_number_300nm" in ser:
            device_point.particle_number_300nm = int(round(ser["particle_number_300nm"]))
        if "sigma_size_dist" in ser:
            device_point.sigma_size_dist = int(round(ser["sigma_size_dist"] * 100.0))
        if "steps_inversion" in ser:
            device_point.steps_inversion = int(round(ser["steps_inversion"]))
        if "current_dist_0" in ser:
            device_point.current_dist_0 = int(round(ser["current_dist_0"] * 100000.0))
        if "current_dist_1" in ser:
            device_point.current_dist_1 = int(round(ser["current_dist_1"] * 100000.0))
        if "current_dist_2" in ser:
            device_point.current_dist_2 = int(round(ser["current_dist_2"] * 100000.0))
        if "current_dist_3" in ser:
            device_point.current_dist_3 = int(round(ser["current_dist_3"] * 100000.0))
        if "current_dist_4" in ser:
            device_point.current_dist_4 = int(round(ser["current_dist_4"] * 100000.0))

        if "supply_voltage_5V" in ser:
            device_point.supply_voltage_5V = int(round(ser["supply_voltage_5V"] * 10.0))
        if "positive_voltage_3V3" in ser:
            device_point.positive_voltage_3V3 = int(round(ser["positive_voltage_3V3"] * 10.0))
        if "negative_voltage_3V3" in ser:
            device_point.negative_voltage_3V3 = int(round(ser["negative_voltage_3V3"] * 10.0))
        if " usb_cc_voltage" in ser:
            device_point.usb_cc_voltage = int(round(ser[" usb_cc_voltage"] * 10.0))

        # Needed for the garagenbox
        if "cs_status" in ser:
            device_point.cs_status = int(ser["cs_status"])

    except Exception as e:
        print(f"Error in _create_device_Point: {e}")
        return None

    return device_point



import logging
from pathlib import Path
from typing import Optional, Union

NANEOS_LOGGER_PATH = "logs/naneos-devices.log"


class CustomFormatter(logging.Formatter):
    _grey = "\x1b[38;20m"
    _green = "\x1b[32;20m"
    _yellow = "\x1b[33;20m"
    _red = "\x1b[31;20m"
    _bold_red = "\x1b[31;1m"
    _reset = "\x1b[0m"
    _format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)"
    _FORMATS_TERMINAL = (
        (logging.DEBUG, f"{_grey} {_format} {_reset}"),
        (logging.INFO, f"{_green} {_format} {_reset}"),
        (logging.WARNING, f"{_yellow} {_format} {_reset}"),
        (logging.ERROR, f"{_red} {_format} {_reset}"),
        (logging.CRITICAL, f"{_bold_red} {_format} {_reset}"),
    )
    _FORMATS_SAVE = (
        (logging.DEBUG, _format),
        (logging.INFO, _format),
        (logging.WARNING, _format),
        (logging.ERROR, _format),
        (logging.CRITICAL, _format),
    )

    def __init__(self, terminal: bool = False, fmt: Optional[str] = None):
        self._FORMATS = self._FORMATS_TERMINAL if terminal else self._FORMATS_SAVE
        if fmt is None:
            fmt = self._format
        super().__init__(fmt=fmt)

    def format(self, record: logging.LogRecord) -> str:
        log_fmt = next(
            (x[1] for x in self._FORMATS if x[0] == record.levelno),
            self._FORMATS[0][1],
        )

        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def set_naneos_logger_save_path(path: Union[str, Path]) -> None:
    global NANEOS_LOGGER_PATH

    if isinstance(path, str):
        path = Path(path).resolve()

    if path.is_dir():
        path = path / "naneos-devices.log"

    # create file if it does not exist
    if not path.exists():
        path.touch()

    NANEOS_LOGGER_PATH = str(path)


def get_naneos_logger(name: str, level: int = logging.INFO) -> logging.Logger:
    global NANEOS_LOGGER_PATH

    logger = logging.getLogger(name)
    logger.setLevel(level)

    # check if NANEOS_LOGGER_PATH exists
    if Path(NANEOS_LOGGER_PATH).exists():
        formatter_file = CustomFormatter(terminal=False)
        file_handler = logging.FileHandler(NANEOS_LOGGER_PATH)
        file_handler.setLevel(level)
        file_handler.setFormatter(formatter_file)
        logger.addHandler(file_handler)

    formatter_stream = CustomFormatter(terminal=True)
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(level)
    stream_handler.setFormatter(formatter_stream)
    stream_handler.terminator = "\r\n"

    logger.addHandler(stream_handler)

    return logger


if __name__ == "__main__":
    set_naneos_logger_save_path("logs/naneos-devices.log")
    logger = get_naneos_logger(__name__, logging.DEBUG)
    logger.debug("debug message")
    logger.info("info message")
    logger.warning("warning message")
    logger.error("error message")
    logger.critical("critical message")



from naneos.iotweb.download.downloader import download_from_iotweb
from naneos.iotweb.naneos_upload_thread import NaneosUploadThread

__all__ = [
    "download_from_iotweb",
    "NaneosUploadThread",
]



import os
import signal
import threading
import time

import pytest

from naneos.manager import NaneosDeviceManager


def raise_keyboard_interrupt():
    os.kill(os.getpid(), signal.SIGINT)


@pytest.mark.timeout(30)
def test_naneos_device_manager():
    timer = threading.Timer(20, raise_keyboard_interrupt)  # trigger keyboard interrupt after 20s
    timer.start()

    manager = NaneosDeviceManager()
    manager.start()

    try:
        while True:
            time.sleep(1)
            print(f"Seconds until next upload: {manager.get_seconds_until_next_upload():.0f}")
            print(manager.get_connected_serial_devices())
            print(manager.get_connected_ble_devices())
            print()
    except KeyboardInterrupt:
        manager.stop()
        manager.join()
        print("NaneosDeviceManager stopped.")

    timer.cancel()



from typing import Optional

import pandas as pd

from naneos.partector.blueprints._data_structure import (
    PARTECTOR1_DATA_STRUCTURE_V_LEGACY,
    NaneosDeviceDataPoint,
)
from naneos.partector.blueprints._partector_blueprint import PartectorBluePrint


class Partector1(PartectorBluePrint):
    def __init__(
        self, serial_number: Optional[int] = None, port: Optional[str] = None, verb_freq: int = 1
    ) -> None:
        super().__init__(serial_number, port, verb_freq)

    def _init_serial_data_structure(self) -> None:
        self.device_type = NaneosDeviceDataPoint.DEV_TYPE_P1
        self._data_structure = PARTECTOR1_DATA_STRUCTURE_V_LEGACY
        self._legacy_data_structure = True

    def _set_verbose_freq(self, freq: int) -> None:
        """
        Set the frequency of the verbose output.

        :param int freq: Frequency of the verbose output in Hz. (0: off, 1: 1Hz, 2: 10Hz, 3: 100Hz)
        """

        if freq < 0 or freq > 3:
            raise ValueError("Frequency must be between 0 and 3!")

        self._write_line(f"X000{freq}!")


if __name__ == "__main__":
    import time

    from naneos.partector.scanPartector import scan_for_serial_partectors

    partectors = scan_for_serial_partectors()
    assert partectors["P1"], "No Partector found!"
    serial_number, port = next(iter(partectors["P1"].items()))

    data: dict[int, pd.DataFrame] = {}
    p1 = Partector1(port=port)

    for _ in range(5):
        time.sleep(5)
        data_points = p1.get_data()
        for point in data_points:
            data = NaneosDeviceDataPoint.add_data_point_to_dict(data, point)

        df = next(iter(data.values()), pd.DataFrame())
        if not df.empty:
            print(f"Sn: {p1._sn}, Port: {p1._port}")
            print(df)
            break

        print("No data received yet...")

    p1.close(blocking=True)



import time
from abc import ABC, abstractmethod
from collections import deque
from datetime import datetime, timezone
from threading import Event, Thread
from typing import Any, Callable, Optional, Union

import serial

from naneos.logger import LEVEL_WARNING, get_naneos_logger
from naneos.partector.blueprints._data_structure import NaneosDeviceDataPoint
from naneos.partector.blueprints._partector_defaults import PartectorDefaults

logger = get_naneos_logger(__name__, LEVEL_WARNING)


class PartectorBluePrint(Thread, PartectorDefaults, ABC):
    """
    Class with the basic functionality of every Partector.
    Mandatory device specific methods are defined abstract and have to be implemented in the child class.
    """

    def __init__(
        self,
        serial_number: Optional[int] = None,
        port: Optional[str] = None,
        verb_freq: int = 1,
        hw_version: str = "None",
    ) -> None:
        """Initializes the Partector2 and starts the reading thread."""
        super().__init__()

        self._init_variables()
        self._verb_freq = verb_freq
        self._init(serial_number, port, verb_freq, hw_version)

    def _init_variables(self) -> None:
        self.device_type: int = 0  # gets initalized by child class
        self._connected = False
        self._sn: Optional[int] = None
        self._port: Optional[str] = None
        self._ser: serial.Serial = serial.Serial()
        self._time_last_message_received = time.time()
        self._legacy_data_structure: bool = False

    #########################################
    ### Init methods
    def _init(
        self,
        serial_number: Optional[int] = None,
        port: Optional[str] = None,
        verb_freq: int = 1,
        hw_version: str = "None",
    ) -> None:
        self._hw_version = hw_version
        self._shutdown_partector = False
        self._init_serial(serial_number, port)
        self._init_thread()
        self._init_data_structures()
        self._init_clear_buffers()
        self.start()

        self._init_get_device_info()

        self._init_serial_data_structure()
        self.set_verbose_freq(verb_freq)

        self._init_print_connection_info()

    def _init_print_connection_info(self) -> None:
        logger.info(f"Connected to SN{self._sn} on {self._port}")

    def _init_serial(self, serial_number: Optional[int] = None, port: Optional[str] = None) -> None:
        self._sn = serial_number
        self._port = port

        self._init_serial_sn_search()
        self._init_serial_connection()

        self._check_connection()

    def _init_serial_sn_search(self) -> None:
        from naneos.partector.scanPartector import scan_for_serial_partector

        if self._sn is not None:
            for _ in range(self.SERIAL_INIT_SCAN_RETRIES):
                self._port = scan_for_serial_partector(self._sn, self._hw_version)
                if self._port:
                    break
            if not self._port:
                logger.warning(
                    f"SN{self._sn} {self._port} not found! Checker is running in the background."
                )
        elif self._port is None:
            raise Exception("No serial number or port given!")

    def _init_serial_connection(self) -> None:
        tries = 0
        tries_start = time.time()

        if self._port is None:
            logger.error("No port given! Cannot initialize serial connection.")
            return

        while time.time() - tries_start < self.SERIAL_INIT_RETRIES_TIMEOUT_S:
            tries += 1

            try:
                self._ser = serial.Serial(
                    port=self._port,
                    baudrate=self.SERIAL_BAUDRATE,
                    timeout=self.SERIAL_TIMEOUT,
                )
            except serial.SerialException:
                continue
            if self._ser.is_open:
                self.set_verbose_freq(0)
                time.sleep(10e-3)
                self._ser.reset_input_buffer()
                break
            self._ser.close()

        if not self._ser.is_open:
            logger.error(f"SN{self._sn} on port {self._port} could not be opened! Tries: {tries}")
        if tries > 1:
            logger.warning(f"SN{self._sn} on port {self._port} needed {tries} tries to connect.")

    def _check_connection(self) -> None:
        if self._ser.is_open:
            self._connected = True
        else:
            self._connected = False
            logger.warning(f"Could not connect to SN{self._sn} on {self._port}")

    def _init_thread(self) -> None:
        self.name = f"naneos-partector-thread_{self._sn}"
        self.thread_event = Event()

    def _init_data_structures(self) -> None:
        self.custom_info_str = "0"
        self.custom_info_size = 0
        # will be declared in child class
        self._data_structure: dict[str, type[Union[int, float]]] = {}
        self._queue: deque[list[Union[int, str]]] = deque(maxlen=self.SERIAL_QUEUE_MAXSIZE)
        self._queue_info: deque[list[Union[int, str]]] = deque(
            maxlen=self.SERIAL_INFO_QUEUE_MAXSIZE
        )

    @abstractmethod
    def _init_serial_data_structure(self) -> None:
        pass

    def _init_clear_buffers(self) -> None:
        if not self._connected:
            return

        time.sleep(10e-3)
        if isinstance(self._ser, serial.Serial):
            self._ser.reset_input_buffer()

    def _init_get_device_info(self) -> None:
        try:
            if self._sn is None:
                self._sn = self._get_serial_number_secure()
            self._fw = self.get_firmware_version()
            logger.debug(f"Connected to SN{self._sn} on {self._port}")
        except Exception:
            logger.warning("Could not get device info!")

    def close(
        self, blocking: bool = True, shutdown: bool = False, verbose_reset: bool = True
    ) -> None:
        """Closes the serial connection and stops the reading thread."""
        self._close(blocking, shutdown, verbose_reset)

    def _checker_thread(self) -> None:
        while not self.thread_event.wait(0.5):
            if time.time() - self._time_last_message_received > 10:
                try:
                    logger.info(f"SN{self._sn} {self._port}: Checking device connection...")
                    self._run_check_connection()
                    self._time_last_message_received = time.time()
                except Exception as e:
                    logger.error(e)

    def _notify_message_received(self) -> None:
        self._time_last_message_received = time.time()

    def run(self) -> None:
        """Thread method. Reads the serial port and puts the data into the queue."""

        # run _checker_thread in own thread
        checker_thread = Thread(target=self._checker_thread)
        checker_thread.start()

        while not self.thread_event.is_set():
            self._run()

        checker_thread.join()

        if self._shutdown_partector:
            self.write_line("off!", 0)

        if isinstance(self._ser, serial.Serial) and self._ser.is_open:
            self._ser.close()

    #########################################
    ### Abstract methods
    def set_verbose_freq(self, freq: int) -> None:
        """
        Sets the verbose frequency of the device.
        This differs for P1, P2 and P2Pro.
        """
        if not self._connected:
            return

        self._set_verbose_freq(freq)

    @abstractmethod
    def _set_verbose_freq(self, freq: int) -> None:
        """
        Sets the verbose frequency of the device.
        This differs for P1, P2 and P2Pro.
        """
        pass

    #########################################
    ### User accessible getters
    def get_serial_number(self) -> Optional[int]:
        """Gets the serial number via command from the device."""
        return self._serial_wrapper(self._get_serial_number)

    def get_firmware_version(self) -> int:
        """Gets the firmware version via command from the device."""
        fw = self._serial_wrapper(self._get_firmware_version)
        if isinstance(fw, int):
            return fw
        return 0

    def write_line(self, line: str, number_of_elem: int = 1) -> list[Any]:
        """
        Writes a custom line to the device and returns the tab-separated response as a list.

        Args:
            line (str): The line to write to the device.
            number_of_elem (int, optional): The number of elements in the response. This will be checked. Defaults to 1.

        Returns:
            list: The response as a list.
        """
        self.custom_info_str = line
        self.custom_info_size = number_of_elem + 1

        if number_of_elem == 0:
            self._write_line(line)
            return []

        return self._serial_wrapper(self._custom_info)  # type: ignore

    #########################################
    ### User accessible data methods
    def clear_data_cache(self) -> None:
        """Clears the data cache."""
        # self._queue.clear()
        if isinstance(self._queue, deque) and len(self._queue) > 0:
            self._queue = deque([self._queue.pop()], maxlen=self.SERIAL_QUEUE_MAXSIZE)

    def get_data(self) -> list[NaneosDeviceDataPoint]:
        points: list[NaneosDeviceDataPoint] = []

        serial_data = list(self._queue)[0:-1]
        self.clear_data_cache()

        for line in serial_data:
            try:
                data_casted = self._cast_splitted_input_string(line)
                point = self._create_naneos_device_point(data_casted)
                points.append(point)
            except Exception as excep:
                logger.warning(f"Could not cast data: {excep}")
                logger.warning(f"Data: {line}")

        return points

    #########################################
    ### Serial methods (private)
    def _close(self, blocking: bool, shutdown: bool, verbose_reset: bool) -> None:
        try:
            if verbose_reset:
                self.set_verbose_freq(0)
                self._write_line("opd00!")
                self._write_line("h2000!")
                self._write_line("e0000!")
        except Exception:
            logger.warning("Could not set verbose frequency to 0!")
        self._shutdown_partector = shutdown
        self.thread_event.set()
        if blocking:
            self.join()

    def _run(self) -> None:
        if not self._connected:
            return

        try:
            if self._ser.is_open:
                self._serial_reading_routine()
        except Exception as e:
            logger.warning(
                f"SN{self._sn} {self._port} Exception occured during threaded serial reading: {e}"
            )

    def _run_check_connection(self) -> bool:
        """Checks if the device is still connected."""
        if not self._connected:
            logger.warning(f"SN{self._sn} {self._port} is not connected! (1)")
            self._init_serial(self._sn, self._port)
            self.set_verbose_freq(self._verb_freq)
            if self._connected:
                self._init_print_connection_info()
        elif self._check_device_connection() is False:
            logger.warning(f"SN{self._sn} {self._port} is not connected! (2)")
            if isinstance(self._ser, serial.Serial) and self._ser.is_open:
                self._ser.close()
            self._connected = False
            self._port = None

        return self._connected

    def _serial_reading_routine(self) -> None:
        if not self._connected:
            return

        line = self._read_line()

        if not line or line == "":
            return

        unix_timestamp = int(datetime.now(tz=timezone.utc).timestamp() * 1000)
        data = [unix_timestamp] + line.split("\t")

        self._notify_message_received()
        if len(data) == len(self._data_structure):
            self._queue.append(data)
        # this is legacy mode, where the data structure is not known exactly
        elif len(data) > len(self._data_structure) and self._legacy_data_structure:
            self._queue.append(data[0 : len(self._data_structure)])
        else:
            self._queue_info.append(data)

    def _check_device_connection(self) -> bool:
        if self.thread_event.is_set() or not self._ser or not self._ser.is_open:
            return False

        try:
            sn = self._get_serial_number_secure()
            if sn == self._sn:
                return True

        except Exception as e:
            logger.error(f"Exception occured during device connection check: {e}")

        return False

    def _check_serial_connection(self) -> None:
        """Tries to reopen a closed connection. Raises exceptions on failure."""
        try:
            for _ in range(3):
                if not self._ser:
                    self._init_serial(self._sn, self._port)
                elif not self._ser.is_open:
                    self._ser.open()

                if self._ser and self._ser.is_open:
                    return None
        except Exception as e:
            if self._ser and self._ser.is_open:
                self._ser.close()
            self._connected = False
            raise ConnectionAbortedError(f"Serial connection aborted: {e}")

    def _serial_wrapper(self, func: Callable[[], Any]) -> Optional[Any]:
        """Wraps user func in try-except block. Forwards exceptions to the user."""
        if not self._connected:
            return None

        logger_msg = ""

        for _ in range(self.SERIAL_RETRIES):
            try:
                return func()
            except Exception as e:
                logger_msg = f"SN{self._sn} Exception occured during user function call: {e}"

        logger.warning(logger_msg)

        return False

    def _write_line(self, line: str) -> None:
        if not self._connected:
            return

        self._check_serial_connection()
        if self._ser:
            self._ser.write(line.encode())
            # time.sleep(10e-3)

    def _read_line(self) -> str:
        if not self._connected:
            return ""

        self._check_serial_connection()
        try:
            data = ""
            if self._ser:
                data = self._ser.readline().decode()
        except Exception as e:
            if self._ser:
                self._ser.close()
            self._connected = False
            raise Exception(f"Was not able to read from the Serial connection: {e}")

        return data.replace("\r", "").replace("\n", "").replace("\x00", "")

    def _get_and_check_info(self, expected_length: int = 2) -> list[Union[int, str]]:
        """
        Get information from the queue and check its length.

        Parameters:
            expected_length (int): The expected length of the information.

        Returns:
            list: The information from the queue.

        Raises:
            ValueError: If the length of the information does not match the expected length.
        """
        # info_data = self._queue_info.get(timeout=self.SERIAL_TIMEOUT_INFO)
        info_data = []
        start_time = time.time()
        while time.time() - start_time < self.SERIAL_TIMEOUT_INFO:
            if len(self._queue_info) > 0:
                info_data = self._queue_info.popleft()
                break
            time.sleep(0.005)
        self._queue_info.clear()

        if len(info_data) != expected_length:
            error_msg = f"Received data of length {len(info_data)}, expected {expected_length}. Data: {info_data}"
            raise ValueError(error_msg)
        return info_data

    def _get_serial_number_secure(self) -> Optional[int]:
        if not self._connected:
            return None

        for _ in range(3):
            serial_numbers = [self.get_serial_number() for _ in range(3)]
            if all(x == serial_numbers[0] for x in serial_numbers):
                return serial_numbers[0]
        raise Exception("Was not able to fetch the serial number (secure)!")

    def _get_serial_number(self) -> int:
        self._queue_info.clear()
        self._write_line("N?")
        return int(self._get_and_check_info()[1])

    def _get_firmware_version(self) -> Optional[int]:
        self._queue_info.clear()
        self._write_line("f?")
        fw = self._get_and_check_info()[1]
        try:
            fw = int(fw)
            return fw
        except Exception as e:
            logger.error(f"Could not cast firmware version to int: {e}")
            return None

    def _custom_info(self) -> list[Union[int, str]]:
        self._queue_info.clear()
        self._write_line(self.custom_info_str)
        return self._get_and_check_info(self.custom_info_size)

    def _cast_splitted_input_string(self, line: list[Union[int, str]]) -> list[Union[int, float]]:
        line_parsed: list[Union[int, float]] = []

        for value, data_type in zip(line, self._data_structure.values()):
            # parsed_value = value if isinstance(value, data_type) else data_type(value)
            parsed_value = data_type(value)

            line_parsed.append(parsed_value)

        return line_parsed

    def _create_naneos_device_point(self, data: list[Union[int, float]]) -> NaneosDeviceDataPoint:
        """
        Creates a NaneosDeviceDataPoint from the given data.

        Args:
            data (list): The data to create the NaneosDeviceDataPoint from.

        Returns:
            NaneosDeviceDataPoint: The created NaneosDeviceDataPoint.
        """
        point = NaneosDeviceDataPoint(
            device_type=self.device_type,
            serial_number=self._sn,
            connection_type=NaneosDeviceDataPoint.CONN_TYPE_SERIAL,
            firmware_version=self._fw,
        )

        for i, name in enumerate(self._data_structure.keys()):
            setattr(point, name, data[i])

        return point



from typing import Optional

from bleak.backends.scanner import AdvertisementData

from naneos.logger import LEVEL_WARNING, get_naneos_logger

logger = get_naneos_logger(__name__, LEVEL_WARNING)


class PartectorBleDecoder:
    """
    Decode the BLE data from the Partector device.
    """

    VALID_DATA_LENGTHS = {22, 44}

    EXPECTED_PROTOCOL_BYTE_1 = "X".encode("utf-8")[0]
    EXPECTED_PROTOCOL_BYTE_1_POSITION = 0
    EXPECTED_PROTOCOL_BYTE_2 = "F".encode("utf-8")[0]
    EXPECTED_PROTOCOL_BYTE_2_POSITION = 21

    EXPECTED_PROTOCOL_BYTE_3 = "Y".encode("utf-8")[0]
    EXPECTED_PROTOCOL_BYTE_3_POSITION = 22
    EXPECTED_PROTOCOL_BYTE_4 = "F".encode("utf-8")[0]
    EXPECTED_PROTOCOL_BYTE_4_POSITION = 43

    SLICE_ADVERTISEMENT = slice(1, 21)
    SLICE_SCAN_RESPONSE = slice(23, 43)

    # == Public Methods ============================================================================
    @classmethod
    def decode_partector_advertisement(
        cls, adv: AdvertisementData
    ) -> Optional[tuple[bytes, Optional[bytes]]]:
        """
        Decode the standard characteristic data from the Partector device.
        """

        adv_bytes = PartectorBleDecoder._get_adv_bytes(adv)
        if not cls._check_data_format(adv_bytes):
            return None

        return cls._remove_protocol_bytes(adv_bytes)

    @staticmethod
    def _get_adv_bytes(adv: AdvertisementData) -> bytes:
        """
        Returns the full advertisement data from the Partector device.
        We are violating the BLE standard here by using the manufacturer data field for our own purposes.
        This is not a good practice, but it was the only way to put more data into the advertisement.
        """
        manufacturer_data = adv.manufacturer_data
        manufacturer_id_bytes = next(iter(manufacturer_data.keys())).to_bytes(2, "little")
        manufacturer_payload = next(iter(manufacturer_data.values()))
        adv_bytes = manufacturer_id_bytes + manufacturer_payload

        return adv_bytes

    @classmethod
    def _check_data_format(cls, data: bytes) -> bool:
        """
        Check if the data format is valid.
        """

        if len(data) not in cls.VALID_DATA_LENGTHS:
            return False

        if (
            data[cls.EXPECTED_PROTOCOL_BYTE_1_POSITION] != cls.EXPECTED_PROTOCOL_BYTE_1
            or data[cls.EXPECTED_PROTOCOL_BYTE_2_POSITION] != cls.EXPECTED_PROTOCOL_BYTE_2
        ):
            return False

        if len(data) > 22:
            if (
                data[cls.EXPECTED_PROTOCOL_BYTE_3_POSITION] != cls.EXPECTED_PROTOCOL_BYTE_3
                or data[cls.EXPECTED_PROTOCOL_BYTE_4_POSITION] != cls.EXPECTED_PROTOCOL_BYTE_4
            ):
                return False

        return True

    @classmethod
    def _remove_protocol_bytes(cls, data: bytes) -> tuple[bytes, Optional[bytes]]:
        """
        Remove the protocol bytes from the data and returns the advertisement data and the scan
        response data in a tuple. The scan response data is optional and may be None if not present.
        """
        adv = data[cls.SLICE_ADVERTISEMENT]
        scan_response = None

        if len(data) > 22:
            scan_response = data[cls.SLICE_SCAN_RESPONSE]

        return (adv, scan_response)



from typing import Optional

from naneos.partector.blueprints._data_structure import NaneosDeviceDataPoint
from naneos.partector_ble.decoder.partector_ble_decoder_blueprint import (
    PartectorBleDecoderBlueprint,
)


class PartectorBleDecoderAux(PartectorBleDecoderBlueprint):
    """
    Decode the std advertisement data from the Partector device.
    """

    OFFSET_CORONA_VOLTAGE = slice(0, 2)
    OFFSET_DIFFUSION_CURRENT = slice(2, 4)
    OFFSET_DEPOSITION_VOLTAGE = slice(4, 6)
    OFFSET_FLOW_FROM_DP = slice(6, 8)
    OFFSET_AMBIENT_PRESSURE = slice(8, 10)
    OFFSET_EM_AMPLITUDE_1 = slice(10, 12)
    OFFSET_EM_AMPLITUDE_2 = slice(12, 14)
    OFFSET_EM_GAIN_1 = slice(14, 16)
    OFFSET_EM_GAIN_2 = slice(16, 18)
    OFFSET_DIFFUSION_CURRENT_OFFSET = slice(18, 20)

    FACTOR_DIFFUSION_CURRENT = 0.01

    # == External used methods =====================================================================
    @classmethod
    def decode(
        cls, data: bytes, data_structure: Optional[NaneosDeviceDataPoint] = None
    ) -> NaneosDeviceDataPoint:
        """
        Decode the auxiliary characteristic data from the Partector device.
        """
        decoded_data = NaneosDeviceDataPoint(
            corona_voltage=cls._get_corona_voltage(data),
            diffusion_current=cls._get_diffusion_current(data),
            deposition_voltage=cls._get_deposition_voltage(data),
            flow_from_dp=cls._get_flow_from_dp(data),
            ambient_pressure=cls._get_ambient_pressure(data),
            electrometer_1_amplitude=cls._get_em_amplitude_1(data),
            electrometer_2_amplitude=cls._get_em_amplitude_2(data),
            electrometer_1_gain=cls._get_em_gain_1(data),
            electrometer_2_gain=cls._get_em_gain_2(data),
            diffusion_current_offset=cls._get_diffusion_current_offset(data),
        )

        if not data_structure:
            return decoded_data

        # Fill the given data_structure with the decoded data
        for field in NaneosDeviceDataPoint.BLE_AUX_FIELD_NAMES:
            setattr(data_structure, field, getattr(decoded_data, field))
        return data_structure

    # == Helpers ===================================================================================
    @classmethod
    def _get_corona_voltage(cls, data: bytes) -> float:
        """
        Get the corona voltage from the advertisement data.
        """
        val = float(int.from_bytes(data[cls.OFFSET_CORONA_VOLTAGE], byteorder="little"))
        return val

    @classmethod
    def _get_diffusion_current(cls, data: bytes) -> float:
        """
        Get the diffusion current from the advertisement data.
        """
        val = float(int.from_bytes(data[cls.OFFSET_DIFFUSION_CURRENT], byteorder="little"))
        return val * cls.FACTOR_DIFFUSION_CURRENT

    @classmethod
    def _get_deposition_voltage(cls, data: bytes) -> float:
        """
        Get the deposition voltage from the advertisement data.
        """
        val = float(int.from_bytes(data[cls.OFFSET_DEPOSITION_VOLTAGE], byteorder="little"))
        return val

    @classmethod
    def _get_flow_from_dp(cls, data: bytes) -> float:
        """
        Get the flow from DP from the advertisement data.
        """
        val = float(int.from_bytes(data[cls.OFFSET_FLOW_FROM_DP], byteorder="little")) / 1000.0
        return val

    @classmethod
    def _get_ambient_pressure(cls, data: bytes) -> float:
        """
        Get the ambient pressure from the advertisement data.
        """
        val = float(int.from_bytes(data[cls.OFFSET_AMBIENT_PRESSURE], byteorder="little"))
        return val

    @classmethod
    def _get_em_amplitude_1(cls, data: bytes) -> float:
        """
        Get the EM amplitude 1 from the advertisement data.
        """
        val = float(int.from_bytes(data[cls.OFFSET_EM_AMPLITUDE_1], byteorder="little"))
        return val

    @classmethod
    def _get_em_amplitude_2(cls, data: bytes) -> float:
        """
        Get the EM amplitude 2 from the advertisement data.
        """
        val = float(int.from_bytes(data[cls.OFFSET_EM_AMPLITUDE_2], byteorder="little"))
        return val

    @classmethod
    def _get_em_gain_1(cls, data: bytes) -> float:
        """
        Get the EM gain 1 from the advertisement data.
        """
        val = float(int.from_bytes(data[cls.OFFSET_EM_GAIN_1], byteorder="little"))
        return val

    @classmethod
    def _get_em_gain_2(cls, data: bytes) -> float:
        """
        Get the EM gain 2 from the advertisement data.
        """
        val = float(int.from_bytes(data[cls.OFFSET_EM_GAIN_2], byteorder="little"))
        return val

    @classmethod
    def _get_diffusion_current_offset(cls, data: bytes) -> float:
        """
        Get the diffusion current offset from the advertisement data.
        """
        val = float(int.from_bytes(data[cls.OFFSET_DIFFUSION_CURRENT_OFFSET], byteorder="little"))
        return val



from dataclasses import dataclass
from typing import Optional, Union

import pandas as pd


def add_to_existing_naneos_data(
    data: dict[int, pd.DataFrame], new_data: dict[int, pd.DataFrame]
) -> dict[int, pd.DataFrame]:
    """
    Adds new data to existing data, merging DataFrames by index.
    If a serial number already exists, it merges the new DataFrame with the existing one.
    """
    if not new_data:  # empty dict
        return data

    for serial, df in list(new_data.items()):
        if serial in data:
            data[serial] = pd.concat([data[serial], df], ignore_index=False)
        else:
            data[serial] = df

    return data


def sort_and_clean_naneos_data(data: dict[int, pd.DataFrame]) -> dict[int, pd.DataFrame]:
    """
    Takes the best connection type for each serial number.
    If there are multiple connection types, it keeps the one with the highest priority:
    - "serial" > "connected" > "advertisement"
    """
    data_return = {}

    for serial, df in data.items():
        if serial is None or df.empty:
            continue

        if "connection_type" in df.columns:
            if (df["connection_type"] == "serial").any():
                df = df[df["connection_type"] == "serial"]
            elif (df["connection_type"] == "connected").any():
                df = df[df["connection_type"] == "connected"]
            elif (df["connection_type"] == "advertisement").any():
                df = df[df["connection_type"] == "advertisement"]

        df = df.sort_index()  # sort by unix_timestamp

        df = df[~df.index.duplicated(keep="last")]

        # if there are rows with device type 2 and 0 remove all 0 rows
        if "device_type" in df.columns:
            if (df["device_type"] == 0).any() and (df["device_type"] == 2).any():
                df = df[df["device_type"] != 0]

        data_return[serial] = df

    return data_return


@dataclass
class NaneosDeviceDataPoint:
    DEV_TYPE_P2 = 0
    DEV_TYPE_P1 = 1
    DEV_TYPE_P2PRO = 2
    DEV_TYPE_P2PRO_CS = 3

    CONN_TYPE_SERIAL = "serial"
    CONN_TYPE_CONNECTED = "connected"
    CONN_TYPE_ADVERTISEMENT = "advertisement"

    BLE_STD_FIELD_NAMES = {
        "serial_number",
        "ldsa",
        "average_particle_diameter",
        "particle_number_concentration",
        "temperature",
        "relative_humidity",
        "device_status",
        "battery_voltage",
        "particle_mass",
    }
    BLE_AUX_FIELD_NAMES = {
        "corona_voltage",
        "diffusion_current",
        "deposition_voltage",
        "flow_from_dp",
        "ambient_pressure",
        "electrometer_1_amplitude",
        "electrometer_2_amplitude",
        "electrometer_1_gain",
        "electrometer_2_gain",
        "diffusion_current_offset",
    }
    BLE_AUX_ERROR_FIELD_NAMES = {
        "device_status",
        "diffusion_current_delay_on",
        "diffusion_current_delay_off",
        "diffusion_current_average",
        "diffusion_current_stddev",
        "diffusion_current_max",
        "corona_voltage_onset",
    }
    BLE_SIZE_DIST_FIELD_NAMES = {
        "particle_number_10nm",
        "particle_number_16nm",
        "particle_number_26nm",
        "particle_number_43nm",
        "particle_number_70nm",
        "particle_number_114nm",
        "particle_number_185nm",
        "particle_number_300nm",
    }

    PANDAS_DTYPES_MAPPING = {
        "unix_timestamp": "Int64",
        "serial_number": "Int32",
        "connection_type": "Int32",
        "firmware_version": "Int32",
        "device_type": "Int32",
        "device_status": "Int32",
        "runtime_min": "Int32",
        "ldsa": "Float32",
        "particle_number_concentration": "Float32",
        "average_particle_diameter": "Float32",
        "particle_mass": "Float32",
        "particle_surface": "Float32",
        "diffusion_current": "Float32",
        "diffusion_current_offset": "Float32",
        "diffusion_current_average": "Float32",
        "diffusion_current_stddev": "Float32",
        "diffusion_current_max": "Float32",
        "diffusion_current_delay_on": "Float32",
        "diffusion_current_delay_off": "Float32",
        "corona_voltage": "Float32",
        "corona_voltage_onset": "Float32",
        "hires_adc1": "Float32",
        "hires_adc2": "Float32",
        "electrometer_1_amplitude": "Float32",
        "electrometer_2_amplitude": "Float32",
        "electrometer_1_gain": "Float32",
        "electrometer_2_gain": "Float32",
        "temperature": "Float32",
        "relative_humidity": "Float32",
        "deposition_voltage": "Float32",
        "battery_voltage": "Float32",
        "flow_from_dp": "Float32",
        "ambient_pressure": "Float32",
        "channel_pressure": "Float32",
        "differential_pressure": "Float32",
        "pump_voltage": "Float32",
        "pump_current": "Float32",
        "pump_pwm": "Float32",
        "particle_number_10nm": "Float32",
        "particle_number_16nm": "Float32",
        "particle_number_26nm": "Float32",
        "particle_number_43nm": "Float32",
        "particle_number_70nm": "Float32",
        "particle_number_114nm": "Float32",
        "particle_number_185nm": "Float32",
        "particle_number_300nm": "Float32",
        "sigma_size_dist": "Float32",
        "steps_inversion": "Float32",
        "current_dist_0": "Float32",
        "current_dist_1": "Float32",
        "current_dist_2": "Float32",
        "current_dist_3": "Float32",
        "current_dist_4": "Float32",
        "supply_voltage_5V": "Float32",
        "positive_voltage_3V3": "Float32",
        "negative_voltage_3V3": "Float32",
        "usb_cc_voltage": "Float32",
        "cs_status": "Int32",
    }

    @staticmethod
    def add_data_point_to_dict(
        devices: dict, data: "NaneosDeviceDataPoint"
    ) -> dict[int, pd.DataFrame]:
        if data.serial_number not in devices:
            devices[data.serial_number] = pd.DataFrame()

        if len(devices[data.serial_number]) > 300:  # remove oldest rows if more than 300 rows
            devices[data.serial_number].drop(devices[data.serial_number].index[0], inplace=True)

        new_row = data.to_pandas_df_row(remove_nan=False)
        if new_row.index.isna():
            return devices

        devices[data.serial_number] = pd.concat(
            [devices[data.serial_number], new_row], ignore_index=False
        )

        return devices

    def to_dict(self, remove_nan=True) -> dict[str, Union[int, float]]:
        if remove_nan:
            return {
                key: getattr(self, key)
                for key in self.__dataclass_fields__
                if getattr(self, key) is not None
            }
        else:
            return {key: getattr(self, key) for key in self.__dataclass_fields__}

    @staticmethod
    def safe_astype(df: pd.DataFrame, mapping: dict[str, str], **kwargs) -> pd.DataFrame:
        mapping = {c: mapping[c] for c in df.columns if c in mapping}
        return df.astype(mapping, **kwargs)

    def to_pandas_df_row(self, remove_nan=True) -> pd.DataFrame:
        """
        Convert the dataclass instance to a pandas DataFrame with a single row.
        """
        data_dict = self.to_dict(remove_nan=remove_nan)
        df = pd.DataFrame([data_dict])
        df = self.safe_astype(df, self.PANDAS_DTYPES_MAPPING, errors="ignore")
        df.set_index(["unix_timestamp"], inplace=True, drop=True)
        return df

    # mandatory
    unix_timestamp: Optional[int] = None
    serial_number: Optional[int] = None
    connection_type: Optional[str] = None  # "serial", "connected", "advertisement"
    firmware_version: Optional[int] = None
    device_type: Optional[int] = 0  # 0: P2, 1: P1, 2: P2PRO, 3: P2PRO_CS
    device_status: Optional[int] = None  # bitmask

    # optional
    runtime_min: Optional[float] = None  # minutes since start
    ldsa: Optional[float] = None  # um**2/cm**3
    particle_number_concentration: Optional[float] = None  # particles/cm**3
    average_particle_diameter: Optional[float] = None  # nm
    particle_mass: Optional[float] = None  # ug/m**3
    particle_surface: Optional[float] = None  # um**2/m**3
    diffusion_current: Optional[float] = None  # nA
    diffusion_current_offset: Optional[float] = None  # nA
    diffusion_current_average: Optional[float] = None  # nA
    diffusion_current_stddev: Optional[float] = None  # nA
    diffusion_current_max: Optional[float] = None  # nA
    diffusion_current_delay_on: Optional[float] = None  # centiseconds
    diffusion_current_delay_off: Optional[float] = None  # centiseconds
    corona_voltage: Optional[float] = None  # V
    corona_voltage_onset: Optional[float] = None  # V
    hires_adc1: Optional[float] = None  # momentanwert em 1
    hires_adc2: Optional[float] = None  # momentanwert em 2
    electrometer_1_amplitude: Optional[float] = None  # mV
    electrometer_2_amplitude: Optional[float] = None  # mV
    electrometer_1_gain: Optional[float] = None  # mV #TODO: check this unit
    electrometer_2_gain: Optional[float] = None  # mV #TODO: check this unit
    temperature: Optional[float] = None  # Celsius
    relative_humidity: Optional[float] = None  # percent 0-100
    deposition_voltage: Optional[float] = None  # V
    battery_voltage: Optional[float] = None  # V
    flow_from_dp: Optional[float] = None  # l/min
    ambient_pressure: Optional[float] = None  # hPa
    channel_pressure: Optional[float] = None  # hPa
    differential_pressure: Optional[float] = None  # Pa
    pump_voltage: Optional[float] = None  # V
    pump_current: Optional[float] = None  # mA
    pump_pwm: Optional[float] = None  # percent 0-100
    particle_number_10nm: Optional[float] = None  # /cm^3/log(d)
    particle_number_16nm: Optional[float] = None  # /cm^3/log(d)
    particle_number_26nm: Optional[float] = None  # /cm^3/log(d)
    particle_number_43nm: Optional[float] = None  # /cm^3/log(d)
    particle_number_70nm: Optional[float] = None  # /cm^3/log(d)
    particle_number_114nm: Optional[float] = None  # /cm^3/log(d)
    particle_number_185nm: Optional[float] = None  # /cm^3/log(d)
    particle_number_300nm: Optional[float] = None  # /cm^3/log(d)
    sigma_size_dist: Optional[float] = None  # gsd
    steps_inversion: Optional[float] = None  # steps count
    current_dist_0: Optional[float] = None  # mV
    current_dist_1: Optional[float] = None  # mV
    current_dist_2: Optional[float] = None  # mV
    current_dist_3: Optional[float] = None  # mV
    current_dist_4: Optional[float] = None  # mV

    supply_voltage_5V: Optional[float] = None  # V
    positive_voltage_3V3: Optional[float] = None  # V
    negative_voltage_3V3: Optional[float] = None  # V
    usb_cc_voltage: Optional[float] = None  # V

    cs_status: Optional[float] = None  # boolean, true or false


PARTECTOR1_DATA_STRUCTURE_V_LEGACY: dict[str, Union[type[int], type[float]]] = {
    "unix_timestamp": int,
    "runtime_min": float,
    "battery_voltage": float,
    "diffusion_current": float,
    "corona_voltage": float,
    "electrometer_1_amplitude": float,  # TODO: check with martin
    "DAC": float,  # TODO: check with martin
    "HVon": int,  # TODO: check with martin
    "idiffset": float,  # TODO: check with martin
    "flow_from_dp": float,
    "ldsa": float,
    "temperature": float,
    "relative_humidity": float,
    "device_status": int,
}

PARTECTOR2_GAIN_TEST_ADDITIONAL_DATA_STRUCTURE: dict[str, Union[type[int], type[float]]] = {
    "electrometer_1_gain": float,
    "electrometer_2_gain": float,
}

PARTECTOR2_OUTPUT_PULSE_DIAGNOSTIC_ADDITIONAL_DATA_STRUCTURE: dict[
    str, Union[type[int], type[float]]
] = {
    "diffusion_current_delay_on": int,  # cs
    "diffusion_current_delay_off": int,  # cs
    "diffusion_current_average": float,  # nA
    "corona_voltage_onset": int,  # V
    "diffusion_current_stddev": float,  # nA
    "diffusion_current_max": float,  # nA
}

PARTECTOR2_DATA_STRUCTURE_V320: dict[str, Union[type[int], type[float]]] = {
    "unix_timestamp": int,
    "runtime_min": float,
    "diffusion_current": float,
    "corona_voltage": int,
    "hires_adc1": float,
    "hires_adc2": float,
    "electrometer_1_amplitude": float,
    "electrometer_2_amplitude": float,
    "temperature": float,
    "relative_humidity": int,
    "device_status": int,
    "deposition_voltage": int,
    "battery_voltage": float,
    "flow_from_dp": float,
    "ldsa": float,
    "average_particle_diameter": float,
    "particle_number_concentration": int,
    "differential_pressure": int,
    "ambient_pressure": float,
}

PARTECTOR2_DATA_STRUCTURE_V295_V297_V298: dict[str, Union[type[int], type[float]]] = {
    "unix_timestamp": int,
    "runtime_min": float,
    "diffusion_current": float,
    "corona_voltage": int,
    "hires_adc1": float,
    "hires_adc2": float,
    "electrometer_1_amplitude": float,
    "electrometer_2_amplitude": float,
    "temperature": float,
    "relative_humidity": int,
    "device_status": int,
    "deposition_voltage": int,
    "battery_voltage": float,
    "flow_from_dp": float,
    "ldsa": float,
    "average_particle_diameter": float,
    "particle_number_concentration": int,
    "differential_pressure": int,
    "ambient_pressure": float,
}

PARTECTOR2_DATA_STRUCTURE_V265_V275: dict[str, Union[type[int], type[float]]] = {
    "unix_timestamp": int,
    "runtime_min": float,
    "diffusion_current": float,
    "corona_voltage": int,
    "hires_adc1": float,
    "hires_adc2": float,
    "electrometer_1_amplitude": float,
    "electrometer_2_amplitude": float,
    "temperature": float,
    "relative_humidity": int,
    "device_status": int,
    "deposition_voltage": int,
    "battery_voltage": float,
    "flow_from_dp": float,
    "ldsa": float,
    "average_particle_diameter": float,
    "particle_number_concentration": int,
    "differential_pressure": int,
    "ambient_pressure": float,
    "lag": int,  # not used anymore
}

PARTECTOR2_DATA_STRUCTURE_LEGACY: dict[str, Union[type[int], type[float]]] = {
    "unix_timestamp": int,
    "runtime_min": float,
    "diffusion_current": float,
    "corona_voltage": int,
    "hires_adc1": float,
    "hires_adc2": float,
    "electrometer_1_amplitude": float,
    "electrometer_2_amplitude": float,
    "temperature": float,
    "relative_humidity": int,
    "device_status": int,
    "deposition_voltage": int,
    "battery_voltage": float,
    "flow_from_dp": float,
    "ldsa": float,
    "average_particle_diameter": float,
    "particle_number_concentration": int,
    "differential_pressure": int,
    "ambient_pressure": float,
}


PARTECTOR2_PRO_DATA_STRUCTURE_V311: dict[str, Union[type[int], type[float]]] = {
    "unix_timestamp": int,
    "runtime_min": float,
    "particle_number_concentration": int,
    "average_particle_diameter": float,
    "ldsa": float,
    "surface": float,  # not existing in protobuf
    "particle_mass": float,
    "sigma_size_dist": float,  # not existing in protobuf
    "diffusion_current": float,
    "corona_voltage": int,
    "deposition_voltage": int,
    "temperature": float,
    "relative_humidity": int,
    "ambient_pressure": float,
    "flow_from_dp": float,
    "battery_voltage": float,
    "pump_current": float,  # not existing in protobuf
    "device_status": int,
    "pump_pwm": int,  # not existing in protobuf
    "steps": int,  # not existing in protobuf
    "particle_number_10nm": int,
    "particle_number_16nm": int,
    "particle_number_26nm": int,
    "particle_number_43nm": int,
    "particle_number_70nm": int,
    "particle_number_114nm": int,
    "particle_number_185nm": int,
    "particle_number_300nm": int,
    "current_dist_0": float,  # not existing in protobuf
    "current_dist_1": float,  # not existing in protobuf
    "current_dist_2": float,  # not existing in protobuf
    "current_dist_3": float,  # not existing in protobuf
    "current_dist_4": float,  # not existing in protobuf
}

PARTECTOR2_PRO_DATA_STRUCTURE_V336: dict[str, Union[type[int], type[float]]] = {
    "unix_timestamp": int,
    "runtime_min": float,
    "particle_number_concentration": int,
    "average_particle_diameter": float,
    "ldsa": float,
    "surface": float,  # not existing in protobuf
    "particle_mass": float,
    "sigma_size_dist": float,  # not existing in protobuf
    "diffusion_current": float,
    "corona_voltage": int,
    "deposition_voltage": int,
    "temperature": float,
    "relative_humidity": int,
    "ambient_pressure": float,
    "flow_from_dp": float,
    "battery_voltage": float,
    "pump_current": float,  # not existing in protobuf
    "device_status": int,
    "flow_from_phase_angle": float,  # not existing in protobuf
    "steps": int,  # not existing in protobuf
    "particle_number_10nm": int,
    "particle_number_16nm": int,
    "particle_number_26nm": int,
    "particle_number_43nm": int,
    "particle_number_70nm": int,
    "particle_number_114nm": int,
    "particle_number_185nm": int,
    "particle_number_300nm": int,
    "current_dist_0": float,
    "current_dist_1": float,
    "current_dist_2": float,
    "current_dist_3": float,
    "current_dist_4": float,
}

PARTECTOR2_PRO_CS_DATA_STRUCTURE_V315: dict[str, Union[type[int], type[float]]] = {
    "unix_timestamp": int,
    "runtime_min": float,
    "particle_number_concentration": int,
    "average_particle_diameter": float,
    "ldsa": float,
    "surface": float,  # not existing in protobuf
    "particle_mass": float,
    "sigma_size_dist": float,  # not existing in protobuf
    "diffusion_current": float,
    "corona_voltage": int,
    "deposition_voltage": int,
    "temperature": float,
    "relative_humidity": int,
    "ambient_pressure": float,
    "flow_from_dp": float,
    "battery_voltage": float,
    "pump_current": float,  # not existing in protobuf
    "device_status": int,
    "pump_pwm": int,  # not existing in protobuf
    "steps": int,  # not existing in protobuf
    "particle_number_10nm": int,
    "particle_number_16nm": int,
    "particle_number_26nm": int,
    "particle_number_43nm": int,
    "particle_number_70nm": int,
    "particle_number_114nm": int,
    "particle_number_185nm": int,
    "particle_number_300nm": int,
    "current_dist_0": float,
    "current_dist_1": float,
    "current_dist_2": float,
    "current_dist_3": float,
    "current_dist_4": float,
    "electrometer_1_gain": float,
    "electrometer_2_gain": float,
    "cs_status": int,
}



from __future__ import annotations

import asyncio
import time

from bleak import BleakScanner
from bleak.backends.device import BLEDevice
from bleak.backends.scanner import AdvertisementData

from naneos.logger import LEVEL_WARNING, get_naneos_logger
from naneos.partector.blueprints._data_structure import NaneosDeviceDataPoint
from naneos.partector_ble.decoder.partector_ble_decoder_aux import PartectorBleDecoderAux
from naneos.partector_ble.decoder.partector_ble_decoder_std import PartectorBleDecoderStd
from naneos.partector_ble.partector_ble_decoder import PartectorBleDecoder

logger = get_naneos_logger(__name__, LEVEL_WARNING)


class PartectorBleScanner:
    """
    Context-managed BLE scanner for Partector devices.

    This scanner runs in the provided asyncio event loop and collects advertisement data
    from BLE devices named "P2" or "PartectorBT". Decoded advertisement payloads are
    pushed into an asyncio.Queue for further processing. Can be used with `async with`
    for automatic startup and cleanup.
    """

    SCAN_INTERVAL = 0.8  # seconds
    BLE_NAMES_NANEOS = {"P2", "PartectorBT"}  # P2 on windows, PartectorBT on linux / mac

    # static methods ###############################################################################
    @staticmethod
    def create_scanner_queue() -> asyncio.Queue[tuple[BLEDevice, NaneosDeviceDataPoint]]:
        """Create a queue for the scanner."""
        queue_scanner: asyncio.Queue[tuple[BLEDevice, NaneosDeviceDataPoint]] = asyncio.Queue(
            maxsize=100
        )

        return queue_scanner

    # == Lifecycle and Context Management ==========================================================
    def __init__(
        self,
        loop: asyncio.AbstractEventLoop,
        queue: asyncio.Queue[tuple[BLEDevice, NaneosDeviceDataPoint]],
    ) -> None:
        """
        Initializes the scanner with the given event loop and queue.

        Args:
            loop (asyncio.AbstractEventLoop): The event loop to run the scanner in.
            queue (asyncio.Queue): The queue to store the scanned data.
        """
        self._loop = loop
        self._queue = queue

        self._task: asyncio.Task | None = None

        self._stop_event = asyncio.Event()
        self._stop_event.set()  # stopped by default

    async def __aenter__(self) -> PartectorBleScanner:
        self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.stop()

    # == Public Methods ============================================================================
    def start(self) -> None:
        """Starts the scanner."""
        if not self._stop_event.is_set():
            logger.warning("You called PartectorBleScanner.start() but scanner is already running.")
            return

        logger.debug("Starting PartectorBleScanner...")
        self._stop_event.clear()
        self._task = self._loop.create_task(self.scan())

    async def stop(self) -> None:
        """Stops the scanner."""
        logger.debug("Stopping PartectorBleScanner...")
        self._stop_event.set()
        if self._task and not self._task.done():
            await self._task
        logger.info("PartectorBleScanner stopped.")

    # == Internal Async Processing =================================================================
    async def _detection_callback(self, device: BLEDevice, adv: AdvertisementData) -> None:
        """Handles the callbacks from the BleakScanner used in the scan method.

        Args:
            device (BLEDevice): Bleak BLEDevice object
            adv (AdvertisementData): Bleak AdvertisementData object
        """

        if not device.name or device.name not in self.BLE_NAMES_NANEOS:
            return

        adv_data = PartectorBleDecoder.decode_partector_advertisement(adv)
        if not adv_data:
            return

        decoded = PartectorBleDecoderStd.decode(adv_data[0], data_structure=None)
        if not decoded.serial_number:
            return
        if adv_data[1]:
            decoded = PartectorBleDecoderAux.decode(adv_data[1], data_structure=decoded)
        decoded.unix_timestamp = int(time.time()) * 1000
        decoded.connection_type = NaneosDeviceDataPoint.CONN_TYPE_ADVERTISEMENT

        if self._queue.full():  # if the queue is full, make space by removing the oldest item
            await self._queue.get()
        await self._queue.put((device, decoded))

    async def scan(self) -> None:
        """Scans for BLE devices and calls the _detection_callback method for each device found."""

        scanner = BleakScanner(self._detection_callback)

        while not self._stop_event.is_set():
            try:
                async with scanner:
                    await asyncio.sleep(self.SCAN_INTERVAL)
            except Exception as e:
                logger.exception(e)
                await asyncio.sleep(self.SCAN_INTERVAL)  # small backoff before retry



import time

from naneos.partector import PartectorSerialManager


def test_readme_example():
    manager = PartectorSerialManager()
    manager.start()

    time.sleep(15)  # Let the manager run for a while
    data = manager.get_data()

    manager.stop()
    manager.join()

    assert isinstance(data, dict), "Data should be a dictionary."
    assert len(data) > 0, "Data dictionary should not be empty."

    print("Collected data:")
    print()
    for sn, df in data.items():
        print(f"SN: {sn}")
        print(df)
        print("-" * 40)
        print()


if __name__ == "__main__":
    test_readme_example()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: protoV1.proto
# Protobuf Python Version: 5.29.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    3,
    '',
    'protoV1.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\rprotoV1.proto\"\xd6\x01\n\x0c\x43ombinedData\x12\x15\n\rabs_timestamp\x18\x01 \x01(\r\x12\x18\n\x07\x64\x65vices\x18\x02 \x03(\x0b\x32\x07.Device\x12\x32\n\x15gateway_points_legacy\x18\x03 \x03(\x0b\x32\x13.GatewayPointLegacy\x12\'\n\x0fposition_points\x18\x04 \x03(\x0b\x32\x0e.PositionPoint\x12\x1f\n\x0bwind_points\x18\x05 \x03(\x0b\x32\n.WindPoint\x12\x17\n\x06uplink\x18\x06 \x01(\x0b\x32\x07.Uplink\"n\n\x06\x44\x65vice\x12\x0c\n\x04type\x18\x01 \x01(\r\x12\x15\n\rserial_number\x18\x02 \x01(\r\x12#\n\rdevice_points\x18\x03 \x03(\x0b\x32\x0c.DevicePoint\x12\x1a\n\x08ui_curve\x18\x04 \x03(\x0b\x32\x08.UiCurve\"\xca\x14\n\x0b\x44\x65vicePoint\x12\x11\n\ttimestamp\x18\x01 \x01(\r\x12\x1a\n\rdevice_status\x18\x02 \x01(\rH\x00\x88\x01\x01\x12\x11\n\x04ldsa\x18\x03 \x01(\rH\x01\x88\x01\x01\x12&\n\x19\x61verage_particle_diameter\x18\x04 \x01(\rH\x02\x88\x01\x01\x12*\n\x1dparticle_number_concentration\x18\x05 \x01(\rH\x03\x88\x01\x01\x12\x18\n\x0btemperature\x18\x06 \x01(\x05H\x04\x88\x01\x01\x12\x1e\n\x11relative_humidity\x18\x07 \x01(\rH\x05\x88\x01\x01\x12\x1c\n\x0f\x62\x61ttery_voltage\x18\x08 \x01(\rH\x06\x88\x01\x01\x12\x1a\n\rparticle_mass\x18\t \x01(\rH\x07\x88\x01\x01\x12\x1b\n\x0e\x63orona_voltage\x18\n \x01(\rH\x08\x88\x01\x01\x12\x1e\n\x11\x64iffusion_current\x18\x0b \x01(\rH\t\x88\x01\x01\x12\x1f\n\x12\x64\x65position_voltage\x18\x0c \x01(\rH\n\x88\x01\x01\x12\x11\n\x04\x66low\x18\r \x01(\rH\x0b\x88\x01\x01\x12\x1d\n\x10\x61mbient_pressure\x18\x0e \x01(\rH\x0c\x88\x01\x01\x12\"\n\x15\x65lectrometer_1_offset\x18\x0f \x01(\x05H\r\x88\x01\x01\x12\"\n\x15\x65lectrometer_2_offset\x18\x10 \x01(\x05H\x0e\x88\x01\x01\x12 \n\x13\x65lectrometer_1_gain\x18\x11 \x01(\rH\x0f\x88\x01\x01\x12 \n\x13\x65lectrometer_2_gain\x18\x12 \x01(\rH\x10\x88\x01\x01\x12%\n\x18\x64iffusion_current_offset\x18\x13 \x01(\x05H\x11\x88\x01\x01\x12!\n\x14particle_number_10nm\x18\x14 \x01(\rH\x12\x88\x01\x01\x12!\n\x14particle_number_16nm\x18\x15 \x01(\rH\x13\x88\x01\x01\x12!\n\x14particle_number_26nm\x18\x16 \x01(\rH\x14\x88\x01\x01\x12!\n\x14particle_number_43nm\x18\x17 \x01(\rH\x15\x88\x01\x01\x12!\n\x14particle_number_70nm\x18\x18 \x01(\rH\x16\x88\x01\x01\x12\"\n\x15particle_number_114nm\x18\x19 \x01(\rH\x17\x88\x01\x01\x12\"\n\x15particle_number_185nm\x18\x1a \x01(\rH\x18\x88\x01\x01\x12\"\n\x15particle_number_300nm\x18\x1b \x01(\rH\x19\x88\x01\x01\x12\x1d\n\x10particle_surface\x18\x1c \x01(\rH\x1a\x88\x01\x01\x12\x1c\n\x0fsigma_size_dist\x18\x1d \x01(\rH\x1b\x88\x01\x01\x12\x1c\n\x0fsteps_inversion\x18\x1e \x01(\rH\x1c\x88\x01\x01\x12\x1b\n\x0e\x63urrent_dist_0\x18\x1f \x01(\rH\x1d\x88\x01\x01\x12\x1b\n\x0e\x63urrent_dist_1\x18  \x01(\rH\x1e\x88\x01\x01\x12\x1b\n\x0e\x63urrent_dist_2\x18! \x01(\rH\x1f\x88\x01\x01\x12\x1b\n\x0e\x63urrent_dist_3\x18\" \x01(\rH \x88\x01\x01\x12\x1b\n\x0e\x63urrent_dist_4\x18# \x01(\rH!\x88\x01\x01\x12\x19\n\x0cpump_current\x18$ \x01(\rH\"\x88\x01\x01\x12\x15\n\x08pump_pwm\x18% \x01(\rH#\x88\x01\x01\x12\x16\n\tcs_status\x18& \x01(\x05H$\x88\x01\x01\x12%\n\x18\x64iffusion_current_stddev\x18\' \x01(\x05H%\x88\x01\x01\x12\'\n\x1a\x64iffusion_current_delay_on\x18( \x01(\rH&\x88\x01\x01\x12(\n\x1b\x64iffusion_current_delay_off\x18) \x01(\rH\'\x88\x01\x01\x12\x19\n\x0cpump_voltage\x18* \x01(\rH(\x88\x01\x01\x12\"\n\x15\x64ifferential_pressure\x18+ \x01(\rH)\x88\x01\x01\x12\x1d\n\x10\x63hannel_pressure\x18, \x01(\rH*\x88\x01\x01\x12\x1e\n\x11supply_voltage_5V\x18- \x01(\rH+\x88\x01\x01\x12!\n\x14positive_voltage_3V3\x18. \x01(\rH,\x88\x01\x01\x12!\n\x14negative_voltage_3V3\x18/ \x01(\rH-\x88\x01\x01\x12\x1b\n\x0eusb_cc_voltage\x18\x30 \x01(\rH.\x88\x01\x01\x12\x1d\n\x10\x66irmware_version\x18\x31 \x01(\rH/\x88\x01\x01\x42\x10\n\x0e_device_statusB\x07\n\x05_ldsaB\x1c\n\x1a_average_particle_diameterB \n\x1e_particle_number_concentrationB\x0e\n\x0c_temperatureB\x14\n\x12_relative_humidityB\x12\n\x10_battery_voltageB\x10\n\x0e_particle_massB\x11\n\x0f_corona_voltageB\x14\n\x12_diffusion_currentB\x15\n\x13_deposition_voltageB\x07\n\x05_flowB\x13\n\x11_ambient_pressureB\x18\n\x16_electrometer_1_offsetB\x18\n\x16_electrometer_2_offsetB\x16\n\x14_electrometer_1_gainB\x16\n\x14_electrometer_2_gainB\x1b\n\x19_diffusion_current_offsetB\x17\n\x15_particle_number_10nmB\x17\n\x15_particle_number_16nmB\x17\n\x15_particle_number_26nmB\x17\n\x15_particle_number_43nmB\x17\n\x15_particle_number_70nmB\x18\n\x16_particle_number_114nmB\x18\n\x16_particle_number_185nmB\x18\n\x16_particle_number_300nmB\x13\n\x11_particle_surfaceB\x12\n\x10_sigma_size_distB\x12\n\x10_steps_inversionB\x11\n\x0f_current_dist_0B\x11\n\x0f_current_dist_1B\x11\n\x0f_current_dist_2B\x11\n\x0f_current_dist_3B\x11\n\x0f_current_dist_4B\x0f\n\r_pump_currentB\x0b\n\t_pump_pwmB\x0c\n\n_cs_statusB\x1b\n\x19_diffusion_current_stddevB\x1d\n\x1b_diffusion_current_delay_onB\x1e\n\x1c_diffusion_current_delay_offB\x0f\n\r_pump_voltageB\x18\n\x16_differential_pressureB\x13\n\x11_channel_pressureB\x14\n\x12_supply_voltage_5VB\x17\n\x15_positive_voltage_3V3B\x17\n\x15_negative_voltage_3V3B\x11\n\x0f_usb_cc_voltageB\x13\n\x11_firmware_version\"@\n\x07UiCurve\x12\x11\n\ttimestamp\x18\x01 \x01(\r\x12\x10\n\x08U_values\x18\x02 \x03(\r\x12\x10\n\x08I_values\x18\x03 \x03(\r\"R\n\x06Uplink\x12\x0c\n\x04type\x18\x01 \x01(\r\x12\x15\n\rserial_number\x18\x02 \x01(\r\x12#\n\ruplink_points\x18\x03 \x03(\x0b\x32\x0c.UplinkPoint\"\x86\t\n\x0bUplinkPoint\x12\x11\n\ttimestamp\x18\x01 \x01(\r\x12\x1d\n\x10\x66irmware_version\x18\x02 \x01(\rH\x00\x88\x01\x01\x12\x1a\n\rdevice_status\x18\x03 \x01(\rH\x01\x88\x01\x01\x12\x1c\n\x0f\x63\x65llular_signal\x18\x04 \x01(\rH\x02\x88\x01\x01\x12 \n\x13\x62\x61ttery_int_voltage\x18\x05 \x01(\rH\x03\x88\x01\x01\x12\x17\n\nvoltage_in\x18\x06 \x01(\rH\x04\x88\x01\x01\x12\x1c\n\x0fusb_voltage_out\x18\x07 \x01(\rH\x05\x88\x01\x01\x12\x1b\n\x0eusb_cc_voltage\x18\x08 \x01(\rH\x06\x88\x01\x01\x12\x1c\n\x0ftemperature_int\x18\t \x01(\x05H\x07\x88\x01\x01\x12\"\n\x15relative_humidity_int\x18\n \x01(\rH\x08\x88\x01\x01\x12\x1e\n\x11temperature_ext_1\x18\x0b \x01(\x05H\t\x88\x01\x01\x12$\n\x17relative_humidity_ext_1\x18\x0c \x01(\rH\n\x88\x01\x01\x12\x1e\n\x11temperature_ext_2\x18\r \x01(\x05H\x0b\x88\x01\x01\x12$\n\x17relative_humidity_ext_2\x18\x0e \x01(\rH\x0c\x88\x01\x01\x12\x15\n\x08sraw_voc\x18\x0f \x01(\rH\r\x88\x01\x01\x12\x16\n\tvoc_index\x18\x10 \x01(\rH\x0e\x88\x01\x01\x12\x1d\n\x10\x61mbient_pressure\x18\x11 \x01(\rH\x0f\x88\x01\x01\x12\x16\n\tco2_ppm_1\x18\x12 \x01(\rH\x10\x88\x01\x01\x12\x16\n\tco2_ppm_2\x18\x13 \x01(\rH\x11\x88\x01\x01\x12\x17\n\npm1_0_mass\x18\x14 \x01(\rH\x12\x88\x01\x01\x12\x17\n\npm2_5_mass\x18\x15 \x01(\rH\x13\x88\x01\x01\x12\x17\n\npm4_0_mass\x18\x16 \x01(\rH\x14\x88\x01\x01\x12\x18\n\x0bpm10_0_mass\x18\x17 \x01(\rH\x15\x88\x01\x01\x12$\n\x17water_level_adc_percent\x18\x18 \x01(\rH\x16\x88\x01\x01\x42\x13\n\x11_firmware_versionB\x10\n\x0e_device_statusB\x12\n\x10_cellular_signalB\x16\n\x14_battery_int_voltageB\r\n\x0b_voltage_inB\x12\n\x10_usb_voltage_outB\x11\n\x0f_usb_cc_voltageB\x12\n\x10_temperature_intB\x18\n\x16_relative_humidity_intB\x14\n\x12_temperature_ext_1B\x1a\n\x18_relative_humidity_ext_1B\x14\n\x12_temperature_ext_2B\x1a\n\x18_relative_humidity_ext_2B\x0b\n\t_sraw_vocB\x0c\n\n_voc_indexB\x13\n\x11_ambient_pressureB\x0c\n\n_co2_ppm_1B\x0c\n\n_co2_ppm_2B\r\n\x0b_pm1_0_massB\r\n\x0b_pm2_5_massB\r\n\x0b_pm4_0_massB\x0e\n\x0c_pm10_0_massB\x1a\n\x18_water_level_adc_percent\"\xba\x02\n\x12GatewayPointLegacy\x12\x16\n\ttimestamp\x18\x01 \x01(\rH\x00\x88\x01\x01\x12\x15\n\rserial_number\x18\x02 \x01(\r\x12\x18\n\x10\x66irmware_version\x18\x03 \x01(\r\x12\x13\n\x0b\x66ree_memory\x18\x04 \x01(\r\x12\x11\n\tfree_heap\x18\x05 \x01(\r\x12\x1f\n\x17largest_free_block_heap\x18\x06 \x01(\r\x12\x17\n\x0f\x63\x65llular_signal\x18\x07 \x01(\x05\x12\x17\n\x0f\x62\x61ttery_int_soc\x18\x08 \x01(\r\x12\x17\n\x0f\x62\x61ttery_ext_soc\x18\t \x01(\r\x12\x1b\n\x13\x62\x61ttery_ext_voltage\x18\n \x01(\r\x12\x1c\n\x14\x63harging_ext_voltage\x18\x0b \x01(\rB\x0c\n\n_timestamp\"Z\n\rPositionPoint\x12\x16\n\ttimestamp\x18\x01 \x01(\rH\x00\x88\x01\x01\x12\x10\n\x08latitude\x18\x02 \x01(\x02\x12\x11\n\tlongitude\x18\x03 \x01(\x02\x42\x0c\n\n_timestamp\"\x81\x01\n\tWindPoint\x12\x16\n\ttimestamp\x18\x01 \x01(\rH\x00\x88\x01\x01\x12\x17\n\nwind_speed\x18\x02 \x01(\rH\x01\x88\x01\x01\x12\x17\n\nwind_angle\x18\x03 \x01(\rH\x02\x88\x01\x01\x42\x0c\n\n_timestampB\r\n\x0b_wind_speedB\r\n\x0b_wind_angleb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'protoV1_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_COMBINEDDATA']._serialized_start=18
  _globals['_COMBINEDDATA']._serialized_end=232
  _globals['_DEVICE']._serialized_start=234
  _globals['_DEVICE']._serialized_end=344
  _globals['_DEVICEPOINT']._serialized_start=347
  _globals['_DEVICEPOINT']._serialized_end=2981
  _globals['_UICURVE']._serialized_start=2983
  _globals['_UICURVE']._serialized_end=3047
  _globals['_UPLINK']._serialized_start=3049
  _globals['_UPLINK']._serialized_end=3131
  _globals['_UPLINKPOINT']._serialized_start=3134
  _globals['_UPLINKPOINT']._serialized_end=4292
  _globals['_GATEWAYPOINTLEGACY']._serialized_start=4295
  _globals['_GATEWAYPOINTLEGACY']._serialized_end=4609
  _globals['_POSITIONPOINT']._serialized_start=4611
  _globals['_POSITIONPOINT']._serialized_end=4701
  _globals['_WINDPOINT']._serialized_start=4704
  _globals['_WINDPOINT']._serialized_end=4833
# @@protoc_insertion_point(module_scope)



import time
import warnings

from naneos.partector import PartectorSerialManager
from naneos.partector.partector1 import Partector1
from naneos.partector.partector2 import Partector2
from naneos.partector.partector2_pro import Partector2Pro
from naneos.partector.scanPartector import scan_for_serial_partectors
from naneos.serial_utils import list_serial_ports


def test_list_serial_ports():
    """
    Test the list_serial_ports function to ensure it returns a list of connected USB devices.
    """
    # Call the function to get the list of serial ports with partectors connected
    serial_ports = list_serial_ports()

    # Check if the result is a list
    assert isinstance(serial_ports, list), "The result should be a list."
    # Check if the list is not empty (assuming there are connected devices)
    assert len(serial_ports) > 0, "There is no connected USB partector device."


def test_connection_partectors() -> None:
    """Test if the serial connection is working 10 times."""
    partectors = scan_for_serial_partectors()
    assert isinstance(partectors, dict), "The result should be a list."
    assert len(partectors) > 0, "There is no connected USB partector device."

    p1 = partectors["P1"]
    p2 = partectors["P2"]
    p2_pro = partectors["P2pro"]

    if len(p1) > 0:
        serial_number = next(iter(p1.keys()))
        for _ in range(5):
            p1 = Partector1(serial_number=serial_number)
            p1.close(verbose_reset=False)
    else:
        warnings.warn("There is no P1 connected (USB).", UserWarning)

    if len(p2) > 0:
        serial_number = next(iter(p2.keys()))
        for _ in range(5):
            p2 = Partector2(serial_number=serial_number)
            p2.close(verbose_reset=False)
    else:
        warnings.warn("There is no P2 connected (USB).", UserWarning)

    if len(p2_pro) > 0:
        serial_number = next(iter(p2_pro.keys()))
        for _ in range(5):
            p2_pro = Partector2Pro(serial_number=serial_number)
            p2_pro.close(verbose_reset=False)
    else:
        warnings.warn("There is no P2pro connected (USB).", UserWarning)


def test_serial_manager():
    manager = PartectorSerialManager()
    manager.start()

    time.sleep(15)  # Let the manager run for a while
    data = manager.get_data()

    manager.stop()
    manager.join()

    assert isinstance(data, dict), "Data should be a dictionary."
    assert len(data) > 0, "Data dictionary should not be empty."

    print("Collected data:")
    print()
    for sn, df in data.items():
        print(f"SN: {sn}")
        print(df)
        print("-" * 40)
        print()



from typing import Optional

import pandas as pd

from naneos.logger import get_naneos_logger
from naneos.partector.blueprints._data_structure import (
    PARTECTOR2_DATA_STRUCTURE_LEGACY,
    PARTECTOR2_DATA_STRUCTURE_V265_V275,
    PARTECTOR2_DATA_STRUCTURE_V295_V297_V298,
    PARTECTOR2_DATA_STRUCTURE_V320,
    PARTECTOR2_GAIN_TEST_ADDITIONAL_DATA_STRUCTURE,
    PARTECTOR2_OUTPUT_PULSE_DIAGNOSTIC_ADDITIONAL_DATA_STRUCTURE,
    NaneosDeviceDataPoint,
)
from naneos.partector.blueprints._partector_blueprint import PartectorBluePrint

logger = get_naneos_logger(__name__)


class Partector2(PartectorBluePrint):
    def __init__(
        self,
        serial_number: Optional[int] = None,
        port: Optional[str] = None,
        verb_freq: int = 1,
        gain_test_active: bool = True,
        output_pulse_diagnostics: bool = True,
    ) -> None:
        self._GAIN_TEST_ACTIVE = gain_test_active
        self._OUTPUT_PULSE_DIAGNOSTICS = output_pulse_diagnostics
        super().__init__(serial_number, port, verb_freq, "P2")

    def _init_serial_data_structure(self) -> None:
        self.device_type = NaneosDeviceDataPoint.DEV_TYPE_P2

        if self._fw in [265, 275]:
            self._data_structure = PARTECTOR2_DATA_STRUCTURE_V265_V275
            logger.info(f"SN{self._sn} has FW{self._fw}. -> Using V265/275 data structure.")
            logger.info("Contact naneos for a firmware update to get the latest features.")
        elif self._fw in [295, 297, 298]:
            self._data_structure = PARTECTOR2_DATA_STRUCTURE_V295_V297_V298
            logger.info(f"SN{self._sn} has FW{self._fw}. -> Using V295/297/298 data structure.")
            logger.info("Contact naneos for a firmware update to get the latest features.")
        elif self._fw >= 320:
            self._data_structure = PARTECTOR2_DATA_STRUCTURE_V320
            self._write_line("A0002!")  # activates antispikes

            if self._OUTPUT_PULSE_DIAGNOSTICS:
                self._write_line("opd01!")
                self._data_structure.update(
                    PARTECTOR2_OUTPUT_PULSE_DIAGNOSTIC_ADDITIONAL_DATA_STRUCTURE
                )
            else:
                self._write_line("opd00!")

            if self._GAIN_TEST_ACTIVE:
                self._write_line("h2001!")  # activates harmonics output
                self._write_line("e1100!")  # strength of gain test signal
                self._data_structure.update(PARTECTOR2_GAIN_TEST_ADDITIONAL_DATA_STRUCTURE)
            else:
                self._write_line("h2000!")  # deactivates harmonics output
                self._write_line("e0000!")  # deactivates gain test signal

            logger.info(f"SN{self._sn} has FW{self._fw}. -> Using V320 data structure.")
        else:
            self._data_structure = PARTECTOR2_DATA_STRUCTURE_LEGACY
            self._legacy_data_structure = True
            logger.warning(f"SN{self._sn} has FW{self._fw}. -> Unofficial firmware version.")
            logger.warning("Using legacy data structure. Contact naneos for a FW update.")

    def _set_verbose_freq(self, freq: int) -> None:
        """
        Set the frequency of the verbose output.

        :param int freq: Frequency of the verbose output in Hz. (0: off, 1: 1Hz, 2: 10Hz, 3: 100Hz)
        """

        if freq < 0 or freq > 3:
            raise ValueError("Frequency must be between 0 and 3!")

        self._write_line(f"X000{freq}!")


if __name__ == "__main__":
    import time

    from naneos.partector.scanPartector import scan_for_serial_partectors

    partectors = scan_for_serial_partectors()
    assert partectors["P2"], "No Partector found!"
    serial_number, port = next(iter(partectors["P2"].items()))

    data: dict[int, pd.DataFrame] = {}
    p2 = Partector2(port=port, gain_test_active=True)

    for _ in range(5):
        time.sleep(3)
        data_points = p2.get_data()
        for point in data_points:
            data = NaneosDeviceDataPoint.add_data_point_to_dict(data, point)

        df = next(iter(data.values()), pd.DataFrame())
        if not df.empty:
            print(f"Sn: {p2._sn}, Port: {p2._port}")
            print(df)
            print(df["diffusion_current_delay_off"])
            data = {}
            # break

    p2.close(blocking=True)



import time

import pandas as pd

from naneos.partector_ble.partector_ble_manager import PartectorBleManager


def test_ble_manager():
    manager = PartectorBleManager()
    manager.start()

    for _ in range(3):
        time.sleep(10)  # Allow some time for the scanner to start
        data = manager.get_data()

        assert isinstance(data, dict)
        assert all(isinstance(df, pd.DataFrame) for df in data.values())

    manager.stop()
